#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Wed Sep 04, 2019 at 06:11 PM -0400

import pytest
import os

from tempfile import NamedTemporaryFile

from pyBabyMaker.babymaker import BabyCppGenerator
from pyBabyMaker.babymaker import BabyMaker

from pyBabyMaker.base import CppCodeDataStore, Variable


PWD = os.path.dirname(os.path.realpath(__file__))
PARDIR = os.path.join(PWD, os.pardir)
SAMPLE_YAML = os.path.join(PARDIR, 'samples', 'sample-babymaker.yml')
SAMPLE_ROOT = os.path.join(PARDIR, 'samples', 'sample.root')
SAMPLE_CPP = os.path.join(PARDIR, 'samples', 'sample_cpp',
                          'sample-babymaker.cpp')


@pytest.fixture
def default_BabyCppGenerator():
    return BabyCppGenerator(list())


def test_default_BabyCppGenerator(default_BabyCppGenerator):
    data_store = CppCodeDataStore(
        input_file='input.root',
        output_file='output.root',
        input_tree='tree_in',
        output_tree='tree_out',
        selection='Y_PT > 1',
        input_br=[Variable('float', 'Y_PT')],
        output_br=[Variable('float', 'y_pt', 'Y_PT+temp')],
        transient=[Variable('float', 'temp', '1')]
    )
    result = default_BabyCppGenerator.gen_preamble_single_output_tree(
        data_store
    )
    assert result == \
        '''
void generator_tree_out(TFile *input_file, TFile *output_file) {
  TTreeReader reader("tree_in", input_file);

  TTree output("tree_out", "tree_out");


  TTreeReaderValue<float> Y_PT(reader, "Y_PT");

  float y_pt_out;
output.Branch("y_pt", &y_pt_out);


  while (reader.Next()) {
    float temp = 1;

    if ((*Y_PT) > 1) {
  y_pt_out = (*Y_PT)+temp;

  output.Fill();
}
  }

  output_file->Write();
}'''


def test_BabyMaker():
    maker = BabyMaker(SAMPLE_YAML, SAMPLE_ROOT, False)

    with open(SAMPLE_CPP, 'r') as f:
        expected = f.read()

    with NamedTemporaryFile() as f:
        maker.write(f.name)
        with open(f.name, 'r') as t:
            generated = t.read()

    assert expected == generated



#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Wed Sep 04, 2019 at 05:48 PM -0400

import os
import pytest
import yaml

from pyBabyMaker.io.NestedYAMLLoader import NestedYAMLLoader

PWD = os.path.dirname(os.path.realpath(__file__))
PARDIR = os.path.join(PWD, os.pardir)
SAMPLE_YAML = os.path.join(PARDIR, 'samples', 'sample-babymaker.yml')


@pytest.fixture
def default_Loader():
    with open(SAMPLE_YAML, 'r') as f:
        return yaml.load(f, NestedYAMLLoader)


def test_NestedYAMLLoader_values(default_Loader):
    result = default_Loader
    assert result['ATuple']['rename'] == {k: k.lower()
                                          for k in ['Y_PT', 'Y_PX', 'Y_PY',
                                                    'Y_PZ']}


def test_NestedYAMLLoader_subfile_values(default_Loader):
    result = default_Loader
    assert result['YetAnotherTuple']['drop'] == ['Y_OWNPV_COV_', 'Y_OWNPV_P.*']



#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Sat Sep 07, 2019 at 11:38 PM -0400

from pyBabyMaker.base import BaseCppGenerator, BaseConfigParser, BaseMaker


class BabyCppGenerator(BaseCppGenerator):
    def gen(self):
        result = ''
        result += self.gen_headers()
        result += self.gen_preamble()
        result += self.gen_body()
        return result

    def gen_preamble(self):
        result = ''
        for data_store in self.instructions:
            result += self.gen_preamble_single_output_tree(data_store)
        return result

    def gen_body(self):
        function_calls = ''.join(
            ['generator_{}(input_file, output_file);\n'.format(
                self.cpp_make_var(i.output_tree)
            ) for i in self.instructions]
        )
        body = '''
TFile *input_file = new TFile(argv[1], "read");
TFile *output_file = new TFile(argv[2], "recreate");

{}

output_file->Close();

delete input_file;
delete output_file;
'''.format(function_calls)
        return self.cpp_main(body)

    def gen_preamble_single_output_tree(self, data_store):
        """
        Generate the body of function call for each ``input_tree`` and
        ``output_tree``.
        """
        input_tree = self.cpp_TTreeReader(
            'reader', data_store.input_tree, 'input_file')
        output_tree = self.cpp_TTree('output', data_store.output_tree)

        input_br = ''.join(
            [self.cpp_TTreeReaderValue(v.type, v.name, 'reader', v.name)
             for v in data_store.input_br])
        output_br = []
        for v in data_store.output_br:
            output_br.append('{} {}_out;\n'.format(v.type, v.name))
            output_br.append('output.Branch("{0}", &{0}_out);\n'.format(v.name))
        output_br = ''.join(output_br)

        transient = ''.join(
            ['{} {} = {};\n'.format(
                v.type, v.name, self.dereference_variables(v.rvalue,
                                                           data_store.input_br))
             for v in data_store.transient])

        output_vars = ''.join(['{}_out = {};\n'.format(
            v.name, self.dereference_variables(v.rvalue, data_store.input_br))
            for v in data_store.output_br])

        if not data_store.selection:
            loop = '''{output_vars}
output.Fill();'''.format(output_vars=output_vars)
        else:
            # We need to prepend a '*' to dereference the value input branches
            loop = '''if ({selection}) {{
  {output_vars}
  output.Fill();
}}'''.format(output_vars=output_vars,
             selection=self.dereference_variables(
                 data_store.selection, data_store.input_br),
             )

        result = '''
void generator_{name}(TFile *input_file, TFile *output_file) {{
  {input_tree}
  {output_tree}

  {input_br}
  {output_br}

  while (reader.Next()) {{
    {transient}
    {loop}
  }}

  output_file->Write();
}}'''.format(name=self.cpp_make_var(data_store.output_tree),
             input_tree=input_tree, output_tree=output_tree,
             input_br=input_br, output_br=output_br,
             transient=transient, loop=loop)

        return result


class BabyMaker(BaseMaker):
    """
    ``babymaker`` class to glue parser and code generator together.
    """
    def __init__(self, config_filename, ntuple_filename, use_reformater=True):
        """
        Initialize with path to YAML file and n-tuple file.
        """
        self.config_filename = config_filename
        self.ntuple_filename = ntuple_filename
        self.use_reformater = use_reformater

    def gen(self):
        parsed_config = self.read(self.config_filename)
        dumped_ntuple = self.dump(self.ntuple_filename)
        parser = self.parse_config(parsed_config, dumped_ntuple)
        generator = BabyCppGenerator(parser.instructions,
                                     parser.system_headers,
                                     parser.user_headers)
        return generator.gen()

    def parse_config(self, parsed_config, dumped_ntuple):
        parser = BaseConfigParser(parsed_config, dumped_ntuple)
        parser.parse()
        return parser

    def write(self, filename):
        content = self.gen()
        with open(filename, 'w') as f:
            f.write(content)
        if self.use_reformater:
            self.reformat(filename)



#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Tue Sep 03, 2019 at 11:55 PM -0400

from pyBabyMaker.parse import is_numeral, find_all_args, find_all_vars


def test_is_numeral_int():
    assert is_numeral(1)


def test_is_numeral_float():
    assert is_numeral(3.14)


def test_is_numeral_str():
    assert is_numeral('3.14')


def test_is_numeral_variable():
    assert not is_numeral('Var3')


def test_find_all_args_simple():
    assert find_all_args('FUNC1(arg1, arg2)') == ['arg1', 'arg2']


def test_find_all_args_nested():
    assert find_all_args(
        'FUNC1(arg1, arg2, Rand(arg3, arg4, FUNC2(arg5 * 2.3, arg6 + arg7)))'
    ) == ['arg1', 'arg2', 'arg3', 'arg4', 'arg5', '2.3', 'arg6', 'arg7']


def test_find_all_args_bool():
    assert find_all_args(
        '!(FUNC1(arg1, arg2) > 1 && FUNC2(FUNC3(arg3, arg4, FUNC4(1, 2)) +'
        'arg6)) <= 3 || FUNC6(arg7 != 3, arg8, arg9*FUNC7())'
    ) == ['arg1', 'arg2', '1', 'arg3', 'arg4', '1', '2', 'arg6', '3', 'arg7',
          '3', 'arg8', 'arg9']


def test_find_all_vars():
    assert find_all_vars(
        '!(FUNC1(arg1, arg2) && FUNC2(FUNC3(arg3, arg4, FUNC4(1, 2)) + arg6))||'
        'FUNC6(arg7, arg8, arg9*FUNC7())'
    ) == ['arg1', 'arg2', 'arg3', 'arg4', 'arg6', 'arg7', 'arg8', 'arg9']



# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Wed Sep 04, 2019 at 11:55 PM -0400

import setuptools
import subprocess
import sys
import re

from setuptools.command.test import test as TestCommand
from distutils.core import setup, Extension

from pyBabyMaker import version


###########
# Helpers #
###########

with open('README.md', 'r') as ld:
    long_description = ld.read()


def get_pipe_output(cmd):
    cmd_splitted = cmd.split(' ')
    proc = subprocess.Popen(cmd_splitted, stdout=subprocess.PIPE)
    result = proc.stdout.read().decode('utf-8')
    return result.strip('\n')


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


########################
# Compile instructions #
########################

root_libdir = get_pipe_output('root-config --libdir')
root_incdir = get_pipe_output('root-config --incdir')

cxx_flags = get_pipe_output('root-config --cflags').split()

# Make sure to enable C++ 14 support
# extra_flags = cxx_flags
extra_flags = [re.sub(r'std=c\+\+11', 'std=c++14', f) for f in cxx_flags]

TupleDumpExtension = Extension(
    name="pyBabyMaker.io.TupleDump",
    sources=["pyBabyMaker/io/TupleDump.cpp"],
    libraries=["RIO", "Tree"],
    library_dirs=[root_libdir],
    include_dirs=[root_incdir],
    extra_compile_args=extra_flags,
    language='c++',
)


#########
# Setup #
#########

setup(
    name='pyBabyMaker',
    version=version,
    author='Yipeng Sun',
    author_email='syp@umd.edu',
    description='Python babymaker (flat ntuple generation tool) library',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/yipengsun/pyBabyMaker',
    packages=setuptools.find_packages(),
    scripts=['bin/ntpdump', 'bin/babymaker'],
    include_package_data=True,
    install_requires=[
        'pyyaml'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent'
    ],
    ext_modules=[TupleDumpExtension],
    tests_require=['pytest'],
    test_suite="test",
    cmdclass={'test': PyTest}
)



#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Sat Sep 07, 2019 at 12:49 PM -0400

name = 'pyBabyMaker'
version = '0.0.1'  # This version is fixed.
version = "0.2.3.3"



#!/usr/bin/env python3
#
# Author: Yipeng Sun <syp at umd dot edu>
# License: BSD 2-clause
# Last Change: Sat Sep 07, 2019 at 11:55 PM -0400
"""
This module provides basic infrastructure for n-tuple related C++ code
generation.
"""

import abc
import yaml
import re
import subprocess

from collections import namedtuple
from datetime import datetime
from shutil import which

from pyBabyMaker.parse import find_all_vars


##################
# Data structure #
##################

class UniqueList(list):
    """
    An extension to the standard ``list`` class such that every element stored
    inside is unique.
    """
    def __init__(self, iterable=None):
        """
        This initializer takes an optional iterable and store the unique
        elements inside that iterable only.
        """
        try:
            uniq = []
            [uniq.append(i) for i in iterable if not uniq.count(i)]
            super().__init__(uniq)
        except TypeError:
            super().__init__()

    def append(self, object):
        if not super().__contains__(object):
            super().append(object)

    def insert(self, index, object):
        if not super().__contains__(object):
            super().insert(index, object)

    def __add__(self, value):
        return UniqueList(super().__add__(value))

    def __iadd__(self, value):
        return UniqueList(super().__iadd__(value))


Variable = namedtuple('Variable', 'type name rvalue, dependency',
                      defaults=(None, None))


class CppCodeDataStore(object):
    """
    Store the data structure for C++ code to be generated.
    """
    def __init__(self, input_file=None, output_file=None,
                 input_tree=None, output_tree=None, selection=None,
                 input_br=None, output_br=None, transient=None):
        """
        Initialize code data store.
        """
        self.input_file = input_file
        self.output_file = output_file
        self.input_tree = input_tree
        self.output_tree = output_tree
        self.selection = selection

        self.input_br = UniqueList(input_br)
        self.output_br = UniqueList(output_br)
        self.transient = UniqueList(transient)

        # Record all loaded variables
        self.loaded_variables = UniqueList()

    def append(self, variable, target):
        """
        Append ``variable`` to ``target``, with the constraint that ``variable``
        must be of ``Variable`` type.
        """
        if type(variable) is not Variable:
            raise TypeError('Type {} is not a valid Variable!'.format(
                type(variable)
            ))
        else:
            self.__getattribute__(target).append(variable)

    def append_input_br(self, variable):
        """
        Append ``variable`` to ``self.input_br``, validating that ``variable``
        has the correct type.

        This will also mark ``variable.name`` as loaded.
        """
        self.append(variable, 'input_br')
        self.loaded_variables.append(variable.name)

    def append_output_br(self, variable):
        """
        Append ``variable`` to ``self.output_br``, validating that ``variable``
        has the correct type.
        """
        self.append(variable, 'output_br')

    def append_transient(self, variable):
        """
        Append ``variable`` to ``self.transient``, validating that ``variable``
        has the correct type.

        This will also mark ``variable.name`` as loaded.
        """
        self.append(variable, 'transient')
        self.loaded_variables.append(variable.name)


###########
# Parsers #
###########

class BaseConfigParser(object):
    """
    Basic parser for YAML C++ code instruction.
    """
    def __init__(self, parsed_config, dumped_ntuple):
        """
        Initialize the config parser with parsed YAML file and dumped n-tuple
        structure.
        """
        self.parsed_config = parsed_config
        self.dumped_ntuple = dumped_ntuple

        self.system_headers = UniqueList()
        self.user_headers = UniqueList()
        self.instructions = []

    def parse(self):
        """
        Parse the loaded YAML dict (in ``self.parsed_config`) and dumped n-tuple
        tree structure (in ``self.dumped_ntuple``).
        """
        for output_tree, config in self.parsed_config.items():
            input_tree = config['input_tree']
            dumped_tree = self.dumped_ntuple[input_tree]
            data_store = CppCodeDataStore(input_tree=input_tree,
                                          output_tree=output_tree)

            self.parse_headers(config)
            self.parse_drop_keep_rename(config, dumped_tree, data_store)
            self.parse_calculation(config, dumped_tree, data_store)
            self.parse_selection(config, dumped_tree, data_store)

            self.instructions.append(data_store)

    def parse_headers(self, config):
        """
        Parse ``headers`` section.
        """
        try:
            self.system_headers += config['headers']['system']
        except KeyError:
            pass

        try:
            self.user_headers += config['headers']['user']
        except KeyError:
            pass

    def parse_drop_keep_rename(self, config, dumped_tree, data_store):
        """
        Parse ``drop, keep, rename`` sections.
        """
        branches_to_keep = []
        for br_in, datatype in dumped_tree.items():
            if 'drop' in config.keys() and self.match(config['drop'], br_in):
                print('Dropping branch: {}'.format(br_in))
            elif 'keep' in config.keys() and self.match(config['keep'], br_in):
                branches_to_keep.append((datatype, br_in))
            elif 'rename' in config.keys() and br_in in config['rename']:
                branches_to_keep.append((datatype, br_in))

        for datatype, br_in in branches_to_keep:
            data_store.append_input_br(Variable(datatype, br_in))
            # Handle branch rename here
            try:
                br_out = config['rename'][br_in]
                data_store.append_output_br(Variable(datatype, br_out, br_in))
            except KeyError:
                data_store.append_output_br(Variable(datatype, br_in, br_in))

    def parse_calculation(self, config, dumped_tree, data_store):
        """
        Parse ``calculation`` section.
        """
        if 'calculation' in config.keys():
            for name, code in config['calculation'].items():
                datatype, rvalue = code.split(';')
                if datatype == '^':
                    self.__getattribute__(rvalue)(name, dumped_tree, data_store)
                elif '^' in datatype:
                    datatype = datatype.strip('^')
                    data_store.append_transient(
                        Variable(datatype, name, rvalue)
                    )
                    self.load_missing_variables(rvalue, dumped_tree, data_store)
                else:
                    data_store.append_output_br(
                        Variable(datatype, name, rvalue)
                    )
                    self.load_missing_variables(rvalue, dumped_tree, data_store)

    def parse_selection(self, config, dumped_tree, data_store):
        """
        Parse ``selection`` section.
        """
        if 'selection' in config.keys():
            data_store.selection = ' '.join(config['selection'])
            self.load_missing_variables(data_store.selection, dumped_tree,
                                        data_store)

    def load_missing_variables(self, expr, dumped_tree, data_store):
        """
        Load missing variables required for calculation or comparison, provided
        that the variables are available directly in the n-tuple.
        """
        variables = find_all_vars(expr)
        for v in variables:
            if v not in data_store.loaded_variables:
                self.LOAD(v, dumped_tree, data_store)

    @staticmethod
    def match(patterns, string, return_value=True):
        """
        Test if ``string`` (a regular expression) matches at least one element
        in the ``patterns``. If there's a match, return ``return_value``.
        """
        for p in patterns:
            if bool(re.search(r'{}'.format(p), string)):
                return return_value
        return not return_value

    @staticmethod
    def LOAD(name, dumped_tree, data_store):
        """
        Load variable ``name`` from n-tuple, if it's available.
        """
        try:
            datatype = dumped_tree[name]
            data_store.append_input_br(
                Variable(datatype, name))
        except KeyError:
            raise KeyError('Branch {} not found.'.format(name))


#######################
# C++ code generators #
#######################


class BaseCppGenerator(metaclass=abc.ABCMeta):
    """
    Basic C++ code snippets for n-tuple processing.
    """
    def __init__(self, instructions,
                 additional_system_headers=None, additional_user_headers=None):
        self.instructions = instructions
        self.system_headers = ['TFile.h', 'TTree.h', 'TTreeReader.h',
                               'TBranch.h']
        self.user_headers = []

        if additional_system_headers is not None:
            self.system_headers += additional_system_headers

        if additional_user_headers is not None:
            self.user_headers += additional_user_headers

    # Code generation ##########################################################

    def gen_headers(self):
        """
        Generate C++ ``#include`` macros.
        """
        system_headers = ''.join([
            self.cpp_header(i) for i in self.system_headers])
        user_headers = ''.join([
            self.cpp_header(i, system=False) for i in self.user_headers])
        return system_headers + '\n' + user_headers

    @abc.abstractmethod
    def gen(self):
        """
        Generate the full C++ output code.
        """

    @abc.abstractmethod
    def gen_preamble(self):
        """
        Generate C++ definitions and functions before ``main``.
        """

    @abc.abstractmethod
    def gen_body(self):
        '''
        Generate C++ code inside ``main`` function.
        '''

    # Helpers ##################################################################

    @staticmethod
    def dereference_variables(expr, vars_to_deref):
        """
        Dereference variables loaded from n-tuple directly. For example:

        .. code-block:: c++

            TTreeReader reader("tree", input_file)
            TTreeReaderValue<double> Y_PT(reader, "Y_PT");

            while (reader.Next()) {
                cout << (*Y_PT)
            }

        The ``Y_PT`` inside the ``while`` loop needs to be dereferenced.
        """
        variables = UniqueList(find_all_vars(expr))
        ref_variables = [v.name for v in vars_to_deref]

        for v in variables:
            if v in ref_variables:
                expr = re.sub(v, '(*{})'.format(v), expr)

        return expr

    # C++ snippets #############################################################

    @staticmethod
    def cpp_gen_date(time_format='%Y-%m-%d %H:%M:%S.%f'):
        """
        C++ code generation time stamp.
        """
        return '// Generated on: {}\n'.format(
            datetime.now().strftime(time_format))

    @staticmethod
    def cpp_header(header, system=True):
        """
        C++ ``#include`` snippets.
        """
        if system:
            return '#include <{}>\n'.format(header)
        else:
            return '#include "{}"\n'.format(header)

    @staticmethod
    def cpp_make_var(name, prefix='', suffix='', separator='_'):
        """
        Make a legal C++ variable name. This is typically used to convert a
        ``TTree`` name to a C++ variable name.
        """
        if prefix != '':
            prefix += separator
        if suffix != '':
            suffix = separator + suffix
        return prefix + re.sub('/', separator, name) + suffix

    @staticmethod
    def cpp_main(body):
        """
        C++ (dumb) main function snippet.
        """
        return '''
int main(int, char** argv) {{
  {0}
  return 0;
}}'''.format(body)

    @staticmethod
    def cpp_TTree(var, name):
        """
        C++ ``TTree`` initializer snippet.
        """
        return 'TTree {0}("{1}", "{1}");\n'.format(var, name)

    @staticmethod
    def cpp_TTreeReader(var, name, TFile):
        """
        C++ ``TTreeReader`` initializer snippet.
        """
        return 'TTreeReader {0}("{1}", {2});\n'.format(var, name, TFile)

    @staticmethod
    def cpp_TTreeReaderValue(datatype, var, TTreeReader, branch_name):
        """
        C++ ``TTreeReaderValue`` initializer snippet.
        """
        return 'TTreeReaderValue<{0}> {1}({2}, "{3}");\n'.format(
            datatype, var, TTreeReader, branch_name
        )


##############
# Base maker #
##############

class BaseMaker(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def parse_config(self, config_filename):
        """
        Parse configuration file for the writer.
        """

    @abc.abstractmethod
    def write(self, filename):
        """
        Write generated C++ file.
        """

    @staticmethod
    def read(yaml_filename):
        """
        Read C++ code generation instruction stored in a YAML.
        """
        from pyBabyMaker.io.NestedYAMLLoader import NestedYAMLLoader
        with open(yaml_filename) as f:
            return yaml.load(f, NestedYAMLLoader)

    @staticmethod
    def dump(data_filename):
        """
        Dump ``TTree`` structures inside a n-tuple
        """
        from pyBabyMaker.io.TupleDump import PyTupleDump
        dumper = PyTupleDump(data_filename)
        return dumper.dump()

    @staticmethod
    def reformat(cpp_filename, formatter='clang-format', flags=['-i']):
        """
        Optionally reformat C++ code after generation, if the ``formatter`` is
        in ``$PATH``.
        """
        if which(formatter):
            cmd = [formatter] + flags + [cpp_filename]
            subprocess.Popen(cmd)
